@@ -1,7 +1,9 @@
+import asyncio
 import datetime
 import logging
 import os
 import shutil
+from functools import partial
 from multiprocessing import cpu_count
 from typing import (
     IO,
@@ -11,15 +13,17 @@ from typing import (
     Dict,
     Iterator,
     List,
+    Literal,
     Optional,
     Union,
     cast,
     overload,
 )
 
+from fsspec.asyn import get_loop
 from funcy import cached_property
 
-from ..executors import ThreadPoolExecutor
+from ..executors import ThreadPoolExecutor, batch_coros
 from .callbacks import DEFAULT_CALLBACK, Callback
 from .errors import RemoteMissingDepsError
 
@@ -27,7 +31,8 @@ if TYPE_CHECKING:
     from typing import BinaryIO, TextIO
 
     from fsspec.spec import AbstractFileSystem
-    from typing_extensions import Literal
+
+    from .path import Path
 
 
 logger = logging.getLogger(__name__)
@@ -68,14 +73,16 @@ class FileSystem:
 
     PARAM_CHECKSUM: ClassVar[Optional[str]] = None
 
-    def __init__(self, **kwargs):
+    def __init__(self, fs=None, **kwargs: Any):
         self._check_requires(**kwargs)
 
         self.jobs = kwargs.get("jobs") or self._JOBS
         self.hash_jobs = kwargs.get("checksum_jobs") or self.HASH_JOBS
         self._config = kwargs
-        self.fs_args = {"skip_instance_cache": True}
+        self.fs_args: Dict[str, Any] = {"skip_instance_cache": True}
         self.fs_args.update(self._prepare_credentials(**kwargs))
+        if fs:
+            self.fs = fs
 
     @property
     def config(self) -> Dict[str, Any]:
@@ -86,7 +93,7 @@ class FileSystem:
         return self.fs.root_marker
 
     @cached_property
-    def path(self):
+    def path(self) -> "Path":
         from .path import Path
 
         def _getcwd():
@@ -102,9 +109,13 @@ class FileSystem:
         return path
 
     @cached_property
-    def fs(self) -> "AbstractFileSystem":
+    def fs(self) -> "AbstractFileSystem":  # pylint: disable=method-hidden
         raise NotImplementedError
 
+    @cached_property
+    def version_aware(self) -> bool:
+        return self.fs_args.get("version_aware", False)
+
     @staticmethod
     def _get_kwargs_from_urls(urlpath: str) -> "Dict[str, Any]":
         from fsspec.utils import infer_storage_options
@@ -164,7 +175,7 @@ class FileSystem:
     def open(
         self,
         path: AnyFSPath,
-        mode: "Literal['rb', 'br', 'wb']",
+        mode: Literal["rb", "br", "wb"],
         **kwargs: Any,
     ) -> "BinaryIO":  # pylint: disable=arguments-differ
         return self.open(path, mode, **kwargs)
@@ -173,7 +184,7 @@ class FileSystem:
     def open(
         self,
         path: AnyFSPath,
-        mode: "Literal['r', 'rt', 'w']",
+        mode: Literal["r", "rt", "w"] = "r",
         **kwargs: Any,
     ) -> "TextIO":  # pylint: disable=arguments-differ
         ...
@@ -201,7 +212,7 @@ class FileSystem:
         self,
         path: Union[AnyFSPath, List[AnyFSPath]],
         recursive: bool = False,
-        on_error: "Literal['raise', 'omit', 'return']" = "raise",
+        on_error: Literal["raise", "omit", "return"] = "raise",
         **kwargs: Any,
     ) -> Union[bytes, Dict[AnyFSPath, bytes]]:
         return self.fs.cat(
@@ -238,6 +249,39 @@ class FileSystem:
     def pipe_file(self, path: AnyFSPath, value: bytes, **kwargs: Any) -> None:
         return self.fs.pipe_file(path, value, **kwargs)
 
+    write_bytes = pipe_file
+    read_bytes = cat_file
+
+    def read_text(
+        self,
+        path: AnyFSPath,
+        encoding: str = None,
+        errors: str = None,
+        newline: str = None,
+        **kwargs: Any,
+    ) -> str:
+        return self.fs.read_text(
+            path, encoding=encoding, errors=errors, newline=newline, **kwargs
+        )
+
+    def write_text(
+        self,
+        path: AnyFSPath,
+        value: str,
+        encoding: str = None,
+        errors: str = None,
+        newline: str = None,
+        **kwargs: Any,
+    ) -> None:
+        self.fs.write_text(
+            path,
+            value,
+            encoding=encoding,
+            errors=errors,
+            newline=newline,
+            **kwargs,
+        )
+
     def pipe(
         self,
         path: Union[AnyFSPath, Dict[AnyFSPath, bytes]],
@@ -263,8 +307,51 @@ class FileSystem:
     ) -> None:
         self.fs.cp_file(from_info, to_info, **kwargs)
 
-    def exists(self, path: AnyFSPath) -> bool:
-        return self.fs.exists(path)
+    @overload
+    def exists(
+        self,
+        path: AnyFSPath,
+        callback: "Callback" = ...,
+        batch_size: Optional[int] = ...,
+    ) -> bool:
+        ...
+
+    @overload
+    def exists(
+        self,
+        path: List[AnyFSPath],
+        callback: "Callback" = ...,
+        batch_size: Optional[int] = ...,
+    ) -> List[bool]:
+        ...
+
+    def exists(
+        self,
+        path: Union[AnyFSPath, List[AnyFSPath]],
+        callback: "Callback" = DEFAULT_CALLBACK,
+        batch_size: Optional[int] = None,
+    ) -> Union[bool, List[bool]]:
+        if isinstance(path, str):
+            return self.fs.exists(path)
+        callback.set_size(len(path))
+        jobs = batch_size or self.jobs
+        if self.fs.async_impl:
+            loop = get_loop()
+            fut = asyncio.run_coroutine_threadsafe(
+                batch_coros(
+                    [
+                        self.fs._exists(p)  # pylint: disable=protected-access
+                        for p in path
+                    ],
+                    batch_size=jobs,
+                    callback=callback,
+                ),
+                loop,
+            )
+            return fut.result()
+        executor = ThreadPoolExecutor(max_workers=jobs, cancel_on_error=True)
+        with executor:
+            return list(executor.map(self.fs.exists, path))
 
     def lexists(self, path: AnyFSPath) -> bool:
         return self.fs.lexists(path)
@@ -275,24 +362,28 @@ class FileSystem:
         except AttributeError:
             raise LinkError("symlink", self, from_info)
 
-    def hardlink(self, from_info: AnyFSPath, to_info: AnyFSPath) -> None:
+    def link(self, from_info: AnyFSPath, to_info: AnyFSPath) -> None:
         try:
-            return self.fs.hardlink(from_info, to_info)
+            return self.fs.link(from_info, to_info)
         except AttributeError:
-            raise LinkError("symlink", self, from_info)
+            raise LinkError("hardlink", self, from_info)
+
+    hardlink = link
 
     def reflink(self, from_info: AnyFSPath, to_info: AnyFSPath) -> None:
         try:
             return self.fs.reflink(from_info, to_info)
         except AttributeError:
-            raise LinkError("symlink", self, from_info)
+            raise LinkError("reflink", self, from_info)
 
-    def is_symlink(self, path: AnyFSPath) -> bool:
+    def islink(self, path: AnyFSPath) -> bool:
         try:
-            return self.fs.is_symlink(path)
+            return self.fs.islink(path)
         except AttributeError:
             return False
 
+    is_symlink = islink
+
     def is_hardlink(self, path: AnyFSPath) -> bool:
         try:
             return self.fs.is_hardlink(path)
@@ -304,23 +395,50 @@ class FileSystem:
 
     @overload
     def ls(
-        self, path: AnyFSPath, detail: "Literal[True]"
+        self, path: AnyFSPath, detail: Literal[True], **kwargs
     ) -> "Iterator[Entry]":
         ...
 
     @overload
-    def ls(self, path: AnyFSPath, detail: "Literal[False]") -> Iterator[str]:
+    def ls(
+        self, path: AnyFSPath, detail: Literal[False], **kwargs
+    ) -> Iterator[str]:
         ...
 
     def ls(self, path, detail=False, **kwargs):
-        return self.fs.ls(path, detail=detail)
+        return self.fs.ls(path, detail=detail, **kwargs)
 
     def find(
         self,
-        path: AnyFSPath,
+        path: Union[AnyFSPath, List[AnyFSPath]],
         prefix: bool = False,  # pylint: disable=unused-argument
+        batch_size: Optional[int] = None,
+        **kwargs,
     ) -> Iterator[str]:
-        yield from self.fs.find(path)
+        if isinstance(path, str):
+            yield from self.fs.find(path)
+            return
+        jobs = batch_size or self.jobs
+        if self.fs.async_impl:
+            loop = get_loop()
+            fut = asyncio.run_coroutine_threadsafe(
+                batch_coros(
+                    [
+                        self.fs._find(p)  # pylint: disable=protected-access
+                        for p in path
+                    ],
+                    batch_size=jobs,
+                ),
+                loop,
+            )
+            for result in fut.result():
+                yield from result
+            return
+        executor = ThreadPoolExecutor(max_workers=jobs, cancel_on_error=True)
+        with executor:
+            find = partial(self.fs.find)
+            for result in executor.imap_unordered(find, path):
+                yield from result
 
     def mv(
         self, from_info: AnyFSPath, to_info: AnyFSPath, **kwargs: Any
@@ -335,13 +453,18 @@ class FileSystem:
     def rm_file(self, path: AnyFSPath) -> None:
         self.fs.rm_file(path)
 
-    def rm(self, path: AnyFSPath) -> None:
-        self.fs.rm(path, recursive=True)
+    def rm(
+        self,
+        path: Union[AnyFSPath, List[AnyFSPath]],
+        recursive: bool = False,
+        **kwargs,
+    ) -> None:
+        self.fs.rm(path, recursive=recursive, **kwargs)
 
     remove = rm
 
-    def info(self, path: AnyFSPath) -> "Entry":
-        return self.fs.info(path)
+    def info(self, path: AnyFSPath, **kwargs) -> "Entry":
+        return self.fs.info(path, **kwargs)
 
     def mkdir(
         self, path: AnyFSPath, create_parents: bool = True, **kwargs: Any
@@ -389,13 +512,8 @@ class FileSystem:
                 length=getattr(fdest, "blocksize", None),  # type: ignore
             )
 
-    def walk(
-        self,
-        path: AnyFSPath,
-        topdown: bool = True,
-        **kwargs: Any,
-    ):
-        return self.fs.walk(path, topdown=topdown, **kwargs)
+    def walk(self, path: AnyFSPath, **kwargs: Any):
+        return self.fs.walk(path, **kwargs)
 
     def glob(self, path: AnyFSPath, **kwargs: Any):
         return self.fs.glob(path, **kwargs)
@@ -453,7 +571,7 @@ class FileSystem:
     ) -> None:
         # Currently, the implementation is non-recursive if the paths are
         # provided as a list, and recursive if it's a single path.
-        from .implementations.local import localfs
+        from .local import localfs
 
         def get_file(rpath, lpath, **kwargs):
             localfs.makedirs(localfs.path.parent(lpath), exist_ok=True)
@@ -528,46 +646,21 @@ class ObjectFileSystem(FileSystem):  # pylint: disable=abstract-method
     ) -> None:
         return None
 
-    def _isdir(self, path: AnyFSPath) -> bool:
-        # Directory in object storages are interpreted differently
-        # among different fsspec providers, so this logic is a temporary
-        # measure for us to adapt as of now. It checks whether it is a
-        # directory (as in a prefix with contents) or whether it is an empty
-        # file where it's name ends with a forward slash
-
-        entry = self.info(path)
-        return entry["type"] == "directory" or (
-            entry["size"] == 0
-            and entry["type"] == "file"
-            and entry["name"].endswith("/")
-        )
-
-    def isdir(self, path: AnyFSPath) -> bool:
-        try:
-            return self._isdir(path)
-        except FileNotFoundError:
-            return False
-
-    def isfile(self, path: AnyFSPath) -> bool:
-        try:
-            return not self._isdir(path)
-        except FileNotFoundError:
-            return False
-
-    def find(self, path: AnyFSPath, prefix: bool = False) -> Iterator[str]:
-        if prefix:
-            with_prefix = self.path.parent(path)
-            files = self.fs.find(with_prefix, prefix=self.path.parts(path)[-1])
-        else:
-            with_prefix = path
-            files = self.fs.find(path)
-
-        # When calling find() on a file, it returns the same file in a list.
-        # For object-based storages, the same behavior applies to empty
-        # directories since they are represented as files. This condition
-        # checks whether we should yield an empty list (if it is an empty
-        # directory) or just yield the file itself.
-        if len(files) == 1 and files[0] == with_prefix and self.isdir(path):
-            return None
-
-        yield from files
+    def find(
+        self,
+        path: Union[AnyFSPath, List[AnyFSPath]],
+        prefix: bool = False,
+        batch_size: Optional[int] = None,  # pylint: disable=unused-argument
+        **kwargs,
+    ) -> Iterator[str]:
+        if isinstance(path, str):
+            paths = [path]
+        for path in paths:
+            if prefix:
+                with_prefix = self.path.parent(path)
+                yield from self.fs.find(
+                    with_prefix, prefix=self.path.parts(path)[-1]
+                )
+            else:
+                with_prefix = path
+                yield from self.fs.find(path)
